package util;

import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.Standardize;

public class Util {
	
	public static float crossValidate(Classifier classifier,Instances data, int n_folds,boolean isRegression) throws Exception {
		 Random rand 		= new Random(0);  	 	// create seeded number generator
		 Instances randData = new Instances(data);  // create copy of original data
		 randData.randomize(rand);         			// randomize data with number generator

		 float performance = 0.0f;
		 for (int n = 0; n < n_folds; n++) {
			   Instances train = randData.trainCV(n_folds, n);
			   Instances test = randData.testCV(n_folds, n);
			
			   classifier.buildClassifier(train);
			   Evaluation eval = new Evaluation(train);
			   eval.evaluateModel(classifier, test);
			   if(isRegression)
				   performance += eval.rootMeanSquaredError();
			   else
				   performance += eval.pctCorrect();
		 }
		 return performance/n_folds;
	}
	public static float crossValidateR(Classifier classifier,Instances data, int n_folds) throws Exception {
		 Random rand 		= new Random(0);  	 	// create seeded number generator
		 Instances randData = new Instances(data);  // create copy of original data
		 randData.randomize(rand);         			// randomize data with number generator

		 float performance = 0.0f;
		 for (int n = 0; n < n_folds; n++) {
			   Instances train = randData.trainCV(n_folds, n);
			   Instances test = randData.testCV(n_folds, n);
			
			   classifier.buildClassifier(train);
			   Evaluation eval = new Evaluation(train);
			   eval.evaluateModel(classifier, test);
			   performance += eval.rootMeanSquaredError();
		 }
		 return 1/(performance/n_folds);
	}
	/**
	 * Return the evaluation value of the classifier, trained on the training instances and tested on the test instances.
	 * 
	 * @param classifier
	 * @param train
	 * @param test
	 * @param isRegression
	 * @return
	 * @throws Exception
	 */
	public static double validateOnce(Classifier classifier, Instances train,Instances test, boolean isRegression) throws Exception {
		classifier.buildClassifier(train);
		Evaluation eval = new Evaluation(train);
		eval.evaluateModel(classifier, test);
		
		if(!isRegression)
			return eval.pctCorrect();
		else
			return 1.0f/eval.rootMeanSquaredError();
	}
	public static Instances copySet(Instances original) throws Exception {
		Instances oracledata = new Instances(original,0);
		for(int i = 0; i < original.size(); i++) {
			Instance proto = original.get(i);
			Instance dense = new DenseInstance(proto.numAttributes());
			dense.setDataset(oracledata);
			for(int a = 0; a < proto.numAttributes(); a++)
				dense.setValue(a, proto.value(a));
			oracledata.add(dense);
		}
		return oracledata;
	}
	/**
	 * Data to [-1,1]
	 */
	public static void normalize(Instances train, Instances test)  {
		Standardize f = new Standardize();

		int Na = train.firstInstance().numAttributes();
		Set<Integer> toNormalize = new TreeSet<Integer>(); 
		double[] maxima = new double[Na];
		double[] minima = new double[Na];		
		//double[] std    = new double[Na];
			
		for(int a = 0; a < Na; a++) {
			if(!train.firstInstance().attribute(a).isNominal()) {
				toNormalize.add(new Integer(a));
				//std[a] = train.variance(a);
			}
		}
		for(int v = 0; v < train.numInstances(); v++) {
			Instance i = train.get(v);
			for(Integer a : toNormalize) {				
				if(maxima[a] < i.value(a)) {
					maxima[a] = i.value(a);
				}
				if(minima[a] > i.value(a)) {
					minima[a] = i.value(a);
				}
			}
		}
		//divide
		for(int v = 0; v < train.numInstances(); v++) {
			Instance i = train.get(v);
			for(Integer a : toNormalize) {				
				i.setValue(a, (i.value(a) - minima[a])/(maxima[a] - minima[a])*2-1);
			}
		}
		for(int v = 0; v < test.numInstances(); v++) {
			Instance i = test.get(v);
			for(Integer a : toNormalize) {				
				i.setValue(a, (i.value(a) - minima[a])/(maxima[a] - minima[a])*2-1);
			}
		}
	}

	private static Instances getPart(Instances data, boolean isTrain, double cutoff) {
		int cutidx = (int) Math.round(data.size()*cutoff);
		Instances splitData = new Instances(data,0);
		
		if(isTrain)
		for(int i = 0 ; i < cutidx; i++) {
			splitData.add((Instance)data.get(i).copy());
		}
		if(!isTrain)
		for(int i = cutidx+1; i < data.size(); i++) {
			splitData.add((Instance)data.get(i).copy());
		}
		return splitData;
	}

}
