package weka.classifiers.rules;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Random;

import util.Util;
import weka.classifiers.Classifier;
import weka.classifiers.trees.J48;
import weka.classifiers.trees.RandomForest;
import weka.core.Capabilities;
import weka.core.CapabilitiesHandler;
import weka.core.DenseInstance;
import weka.core.Drawable;
import weka.core.Instance;
import weka.core.Instances;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.WoE;

@SuppressWarnings("serial")
public class ALPAC implements Classifier,Serializable, CapabilitiesHandler,Drawable {
	final int DEBUG_LEVEL = 0;
	double	val_fidelity = 0.0f;
	double 	val_rho 	 = 0.0f;
	
	/******************************* PUBLIC OPTIONS *******************************/		
	private float n_valleypoints = 0.25f; //Percentage of valley points	
	private float rho_interval 	 = 0.05f;
	private boolean isRegression = false;
	
	public Classifier whiteBox,blackBox,alpacBox; //TODO: make private	
	
	public float getN_valleypoints() {
		return n_valleypoints;
	}
	public void setN_valleypoints(float n_valleypoints) {
		this.n_valleypoints = n_valleypoints;
	}
	public float getRho_interval() {
		return rho_interval;
	}
	public void setRho_interval(float rho_interval) {
		this.rho_interval = rho_interval;
	}
	public Classifier getWhiteBox() {
		return whiteBox;
	}
	public void setWhiteBox(Classifier whiteBox) {
		this.whiteBox = whiteBox;
	}
	public Classifier getBlackBox() {
		return blackBox;
	}
	public void setBlackBox(Classifier blackBox) {
		this.blackBox = blackBox;
	}
	
	/******************************* HELPER CLASSES *******************************/
	//Helper class for sorting out valley points based on indices instead of values
	public class ArrayIndexComparator<T extends Comparable<T>> implements Comparator<Integer> {
	    private final Object[] array;

	    public ArrayIndexComparator(Object[] array) {
	        this.array = array;
	    }
	    public Integer[] createIndexArray() {
	        Integer[] indexes = new Integer[array.length];
	        for (int i = 0; i < array.length; i++) {
	            indexes[i] = i; // Autoboxing
	        }
	        return indexes;
	    }
	    @Override 
	    public int compare(Integer index1, Integer index2) { //TODO: warning due to ugly cast
	        return ((T)array[index1]).compareTo((T)array[index2]);
	    }
	}

	/******************************* CLASSIFIER *******************************/	
	Instances m_Instances;
	int m_nAttributes;
	
	public ALPAC() {
		this(new J48(),new RandomForest());
	}
	public ALPAC(Classifier whiteBox, Classifier blackBox) {
		this.whiteBox = whiteBox;
		this.blackBox = blackBox;
	}
	public void buildClassifier(Instances data) throws Exception {
		reset();
		getCapabilities().testWithFail(data);
		 try {
			 alpacBox = ALPA(data);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	public void reset() {
		valleypoints= null;
		nearest		= null;
		alpacBox 	= null;
	}
	@SuppressWarnings("unused")
	public Classifier ALPA(Instances data) throws Exception {
		if(DEBUG_LEVEL > 0) System.out.print("ALPA ");
		if(data == null || data.size() <= 0)
			throw new Exception("Size of dataset must be greater than zero.");
		
		//1. Build the oracle set: using predictions instead of true class values		
		int Nt = data.size();
		blackBox.buildClassifier(data);
		
		//2. For all rhos:		
		double bestPerf = 0;
		float  bestRho = 0;
		for(float rho = 0.0f; rho < 2.5f; rho += rho_interval) {
			if(DEBUG_LEVEL > 0) System.out.print(".");
			//System.out.print(rho+":");
			//a) Generate extra points
			int Nv = (int) Math.round(n_valleypoints*Nt);
			int Nr = (int) Math.round(rho*Nt);	
			//if(Nr < 1) Nr = 1;
						
			Instances artificial = generatePoints(blackBox, data, Nv, Nr);
			Instances oracledata = getOracleData(data, blackBox);

			//b) Add artificial points to dataset
			Instances augmented = new Instances(oracledata,0);			
			augmented.addAll(oracledata);
			augmented.addAll(artificial);
			
			// c) train & evaluate model
			Instances train 	= 	augmented.trainCV(3,0);
			Instances test 		= 	augmented.testCV(3,0);
			whiteBox.buildClassifier(train);
			double perf = Util.validateOnce(whiteBox, train,test,isRegression);
			
			//TODO: should do more than one test, cross validate?
			// d) is it better? keep it!
			 if(perf > bestPerf) {
				 bestPerf 		= perf;
				 bestRho 		= rho;
			 }
		}
		val_fidelity 	= bestPerf;
		val_rho 		= bestRho;
		
		if(DEBUG_LEVEL > 0) System.out.println("Best performance "+bestPerf+ " at "+bestRho);
		if(DEBUG_LEVEL > 0) System.out.println();
		
		//Retrain best model
		int Nv = (int) Math.round(n_valleypoints*Nt);
		int Nr = (int) Math.round(bestRho*Nt);	
		if(Nr < 1) Nr = 1;		
		blackBox.buildClassifier(data);			
		Instances artificial = generatePoints(blackBox, data, Nv, Nr);
		Instances oracledata = getOracleData(data, blackBox);
		Instances augmented = new Instances(oracledata,0);			
		augmented.addAll(oracledata);
		augmented.addAll(artificial);
		Instances train 	= 	augmented.trainCV(3,0);
		Instances test 		= 	augmented.testCV(3,0);
		whiteBox.buildClassifier(train);
		
		double perf = Util.validateOnce(whiteBox, train,test,isRegression);
		//System.out.println(perf);

		return whiteBox;
	}

	private Instances getOracleData(Instances original, Classifier blackBox) throws Exception {
		Instances oracledata = new Instances(original,0);
		for(int i = 0; i < original.size(); i++) {
			Instance proto = original.get(i);
			Instance dense = new DenseInstance(proto.numAttributes());
			dense.setDataset(oracledata);
			for(int a = 0; a < proto.numAttributes(); a++)
				dense.setValue(a, proto.value(a));
			dense.setClassValue(blackBox.classifyInstance(dense));
			oracledata.add(dense);
		}
		return oracledata;
	}
	
	private Instances valleypoints; 
	private int[] nearest; 		

	public Instances generatePoints(Classifier blackBox, Instances data, int Nv, int Nr) throws Exception {
		if(nearest == null) {
			List<Double> scores = new ArrayList<Double>();

			//1. Create valleypoints set
			for(Instance i : data) {
				double score = 0;
				if(isRegression) {
					double y_hat = blackBox.classifyInstance(i);
					double y 	 = i.classValue();
					score = Math.abs(y_hat-y);
				}
				else {
					double[] distr =  blackBox.distributionForInstance(i);
					score =  2*distr[(int) blackBox.classifyInstance(i)];
					for(int c = 0; c < i.numClasses(); c++)
						score -= distr[c];
				}
				scores.add(score);
			}
			ArrayIndexComparator<Float> comparator = new ArrayIndexComparator<Float>(scores.toArray());
			Integer[] indexes = comparator.createIndexArray(); //indexen van scores van klein naar groot
			Arrays.sort(indexes, comparator);
						
			valleypoints = new Instances(data,0);			
			for(int i = 0; i < Nv; i++) {
				valleypoints.add(data.get(indexes[i]));
			}

			//calculate distances for this valleypoint set - nominals are converted with WoE
			WoE filterWoE = new WoE();
			filterWoE.setInputFormat(valleypoints);   
			Instances realData = Filter.useFilter(valleypoints, filterWoE);
			realData.setClassIndex(valleypoints.classIndex());
			float[][] distances = getDistances(realData);			

			//calculate the nearest neighbout for each of the valleypoints
			nearest = getNearest(valleypoints,distances);
		}

		//TODO: De eerste transformatie moet in principe niet altijd, kan ook opgeslagen worden
		//a) 	transform the vallypoints into real-representation
		WoE filterWoE = new WoE();
		filterWoE.setInputFormat(valleypoints);   
		Instances realData = Filter.useFilter(valleypoints, filterWoE);
		realData.setClassIndex(valleypoints.classIndex());
		
		//b) 	generate new points
		Instance prototype = realData.firstInstance();
		int Na = prototype.numAttributes();
		
		Instances randomData = new Instances(valleypoints,0);
		Random generator = new Random(0);
		for(int r = 0; r < Nr; r++) {
			int i1 = generator.nextInt(Nv);
			int i2 = nearest[i1];
			Instance v1 = valleypoints.get(i1);
			Instance v2 = valleypoints.get(i2);
			
			double theta = generator.nextDouble();
			
			Instance instance = new DenseInstance(prototype.numAttributes()); 
			instance.setDataset(randomData);			
			for(int a = 0; a < Na; a++) {
				double mix = theta*v1.value(a) + (1-theta)*v2.value(a);
				instance.setValue(prototype.attribute(a), mix);
			}
			randomData.add(instance);
		}
		//c)	transform all points back to nominal-representation 
		filterWoE.setInverseMode(true);
		filterWoE.setInputFormat(valleypoints);   
		Instances nominalData = Filter.useFilter(randomData, filterWoE);
		nominalData.setClassIndex(data.classIndex());
		
		//d) 	classify them (black-label)
		for(Instance instance : nominalData)
			instance.setClassValue(blackBox.classifyInstance(instance));

		return nominalData;
	}
	/**
	 * 
	 * @param valleypoints2 	a set of previously defined valley points
	 * @param distances 		the distances between each of these valley points
	 * @return 					returns an array where position i contains the index of the nearest valleypoint for valleypoint_i
	 */
	private int[] getNearest(Instances valleypoints2, float[][] distances) {
		int Nv = valleypoints2.size();
		int[] nearest = new int[Nv];	
		
		for(int v1 = 0; v1 < Nv; v1++) {
			float min   = Float.MAX_VALUE;
			int minv    = v1;
			for(int v2 = 0; v2 < Nv; v2++) {
				if(distances[v1][v2]  < min && v1 != v2) {
					min  = distances[v1][v2];
					minv = v2;
				}
			}
			nearest[v1] = minv;
		}
		return nearest;
	}
	public float[][] getDistances(Instances lValleypoints) {
		int Nv = lValleypoints.size();
		int Na = lValleypoints.firstInstance().numAttributes();
		

		float[][] distances = new float[Nv][Nv];
		for(int v1 = 0; v1 < Nv; v1++) {
			Instance vp1 = lValleypoints.get(v1);
			for(int v2 = v1; v2 < Nv; v2++) {					
				Instance vp2 = lValleypoints.get(v2);								
				float d = 0;
				if(vp1.classValue() != vp2.classValue()) {					
					for(int a = 0; a < Na ; a ++) {
						double x1 = vp1.value(a);
						double x2 = vp2.value(a);
						d += (x1 - x2)*(x1 - x2);
					}
				} else {
					d = Float.MAX_VALUE;
				}
				distances[v1][v2] = d;
			}
			for(int v2 = 0; v2 < v1; v2++)
				distances[v1][v2] = distances[v2][v1];
		}
		return distances;
	}
	
	@Override
	public double classifyInstance(Instance instance) {
		if(alpacBox == null) System.err.println("You must train the classifier first!");
		try {
			return alpacBox.classifyInstance(instance);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return 0;
	}
	@Override
	public double[] distributionForInstance(Instance arg0) throws Exception {
		if(alpacBox == null) {
			System.err.println("You must train the classifier first!");
			return new double[]{};
		}
		return alpacBox.distributionForInstance(arg0);
	}

	//this function is used to show the list of the parameter options
	public String [] getOptions() {
		int current = 0;
		String [] options = new String [2];
		while (current < options.length)
			options[current++] = "";
		return options;
	}
	@Override
	public Capabilities getCapabilities() {
		Capabilities c = new Capabilities(this);
		c.or(blackBox.getCapabilities());
		c.and(whiteBox.getCapabilities());
	    return c;
	}
	public static String printPercentage(double pc) {
		return ""+Math.round(pc*100)/100.0f;
	}
	/*** DRAWABLE ***/
	@Override
	public String graph() throws Exception {
		if(whiteBox instanceof Drawable) {
			return ((Drawable) whiteBox).graph();
		}
		return null;
	}
	@Override
	public int graphType() {
		if(whiteBox instanceof Drawable) {
			return ((Drawable) whiteBox).graphType();
		}
		return 0;
	}
	public String toString() {
		String out 	=  "Active Learning-based Pedagogical Rule Extraction\n\n";
		out 		+= "Validation fidelity : "+fiveDigits(val_fidelity)+"% \n";
		out 		+= "Validation rho      : "+fiveDigits(val_rho*100) +"%\n";
		out 		+= "\n WhiteBox model:\n\n"+whiteBox.toString();
		return out;
	}
	public String fiveDigits(double f) {
		f = f*Math.pow(10,5);
		f = Math.round(f);
		f = f/Math.pow(10,5);
		return ""+f;
	}
}