package weka.filters.unsupervised.attribute;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import weka.core.Attribute;
import weka.core.Capabilities;
import weka.core.Capabilities.Capability;
import weka.core.DenseInstance;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.filters.SimpleBatchFilter;
/**
 * Usage:
 * 		woe.setInputFormat(trainset);
 *  	Filter.useFilter(data, woe);
 *  
 *  Afterwards:
 *     	woe2.setInputFormat(data);   
 *  	woe2.setInverseMode(true);
 *  	Filter.useFilter(data, woe2);
 *  
 * @author ejdfortu
 *
 */
@SuppressWarnings({ "serial", "deprecation" })
public class WoE extends SimpleBatchFilter {
	boolean	verbose		= false;
	boolean inverseMode = false;
	boolean isRegression= true;
	
	Instances nominalPrototype;
	Instances woePrototype;
	
	public WoE() {}
	public WoE(boolean isRegression) {
		this.isRegression = isRegression;
	}
	public String globalInfo() {
		return "A simple batch filter that converts all nominal values to their respective weight of evidence values.";
	}

	public Capabilities getCapabilities() {
		Capabilities result = super.getCapabilities();
		result.enableAllAttributes();
		result.enableAllClasses();
		result.enable(Capability.NO_CLASS); // // filter doesn't need class to
											// be set//
		return result;
	}
	public void setInverseMode(boolean inverseMode) {
		this.inverseMode = inverseMode;
	}
	Map<Integer, Map<Object, Integer>>	evidenceMap;	//Bevat voor elk attribuut de hoeveelheid in dataset met dit attribuut
	Map<Integer, Map<Object,Float>> 	nomToRealMap;	//Bevat voor elk attribuut, voor elke waarde naar welk real die gemapt moet worden
	Map<Integer,Map<Float,Object>> 		realToNomMap; 	//Bevat voor elk attribuut, voor elke real naar welk oorspronkelijke nominale waarde die gemapt moet worden

	@Override
	public boolean setInputFormat(Instances instances) {
		//Preliminary checks and processing
		try {
			super.setInputFormat(instances);
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}		
		if(instances.size() <= 0) return false;
		
		evidenceMap 	= new HashMap<Integer,Map<Object,Integer>>();
		nomToRealMap 	= new HashMap<Integer, Map<Object,Float>>();
		realToNomMap 	= new HashMap<Integer,Map<Float,Object>>();
		nominalPrototype= new Instances(instances,0);
		nominalPrototype.add(new DenseInstance(instances.firstInstance().numAttributes()));
		
		//Check for nominal attributes
		if(verbose)
			System.out.print("setInputFormat:");

		// 1. Determine all the possible attributes values
		Instance prototype = instances.firstInstance();
		List<Integer> nominals = new ArrayList<Integer>();
		for(int a = 0; a < prototype.numAttributes(); a++) {
			if(prototype.attribute(a).isNominal()) {
				nominals.add(a);
			}
		}


		// 2. Determine presence amount of each value in the attribute map
		for(Integer index : nominals) {
			Map<Object,Integer> valueMap = new HashMap<Object,Integer>();
			
			//CLASSIFICATION : actual presence
			if(!isRegression)
			for(Instance i : instances) {
				Object value = i.value(i.attribute(index));
				if(!valueMap.containsKey(value))
					valueMap.put(value, new Integer(1));
				else
					valueMap.put(value, valueMap.get(value)+1);
			}
			//REGRESSION : each value counts equally
			if(isRegression)
			for(Instance i : instances) {
				Object value = i.value(i.attribute(index));
				if(!valueMap.containsKey(value))
					valueMap.put(value, new Integer(1));
			}
			
			
			evidenceMap.put(index,valueMap);
			if(verbose)
				System.out.println("\tAttribute "+index+" has " + valueMap.size()+ " possible values.");

			//3. Define scale mapping values
			Map<Object,Float> nomToReal = new HashMap<Object,Float>();
			Map<Float,Object> realToNom = new HashMap<Float,Object>();

			float maxSupport = instances.size();
			float relative = 0;
			for(Entry<Object, Integer> entry : valueMap.entrySet()) {
				float mySupport 	= (entry.getValue()*1.0f) / maxSupport;
				float otherSupport 	= (maxSupport - entry.getValue()) / maxSupport;				
				float WoE = (float) ((mySupport / otherSupport) / Math.log(Math.E));
				relative += WoE;				
				nomToReal.put(entry.getKey(), new Float(relative));
				realToNom.put(new Float(relative),entry.getKey());
			}
			nomToRealMap.put(index,nomToReal);
			realToNomMap.put(index, realToNom);
		}
		// 4. Map(attribute).put: NOMINAL_VALUE -> WoE VALUE		
		if(verbose)
			System.out.println("detected "+nomToRealMap.keySet().size()+" nominal features.");
		
		//5. Create a new prototype for the real-valued instance
	    FastVector<Attribute> atts = new FastVector<Attribute>();
	    for(int a = 0; a < prototype.numAttributes(); a++) {
	     Attribute att = prototype.attribute(a);
	     	if(nomToRealMap.containsKey(a))
	    	 	att = new Attribute("WoE "+a);    		 
	     	atts.add(att);
	    }
	     woePrototype = new Instances(nominalPrototype.relationName()+" (WoE)", atts, 1);
	     woePrototype.add(new DenseInstance(atts.size()));
	     return true;
	}
	protected Instances determineOutputFormat(Instances inputFormat) {
		if(inverseMode)
			return nominalPrototype;
		else
			return woePrototype;
	}
	protected Instances process(Instances inst) {
		if(inverseMode)
			return processToNominal(inst);
		else
			return processToWoE(inst);

	}
	//Van nominaal naar real vertalen
	private Instances processToWoE(Instances inst) {
		Instances newInst = new Instances(woePrototype,0);		
		for(Instance i : inst) {
			Instance newInstance = new DenseInstance(woePrototype.firstInstance());
			for(int a = 0; a < i.numAttributes(); a++) {				
				if(i.attribute(a).isNominal()) {					
					Float newValue 	= nomToRealMap.get(a).get(i.value(a));
					newInstance.setValue(a, newValue);				
				}
				else
					newInstance.setValue(a,i.value(a));
			}
			newInst.add(newInstance);
		}				
		return newInst;		
	}
	//Van real naar nominaal vertalen
	private Instances processToNominal(Instances inst) {
		Instances newInst = new Instances(nominalPrototype,0);		
		for(Instance i : inst) {
			Instance newInstance = new DenseInstance(nominalPrototype.firstInstance());
			newInstance.setDataset(nominalPrototype);
			for(int a = 0; a < i.numAttributes(); a++) {				
				if(nominalPrototype.attribute(a).isNominal()) {					
					Float nearest 	= findNearest(a,i.value(a));
					Object newValue = realToNomMap.get(a).get(nearest);
					newInstance.setValue(a, (double) newValue);				
				}
				else
					newInstance.setValue(a,i.value(a));
			}
			newInst.add(newInstance);
		}				
		return newInst;		
	}
	//TODO: voor nominalen met heel veel waarden is dit traag -> Log2 Search
	public Float findNearest(int attribute, double real) {
		Map<Float, Object> map = realToNomMap.get(attribute);
		
		List<Float> collection = new ArrayList<Float>();
		collection.addAll(map.keySet());		
		Collections.sort(collection);		
		Float[] values = collection.toArray(new Float[]{});
		
		int i=0;
		while(values[i] <= real && i < values.length-1) {
			i++;
		}
		
		if( i <= 0) 						return values[0];
		if( i == values.length) 			return values[i-1];
		
		float rightVal = values[i];
		float leftVal  = values[i-1];
		
		if( real-leftVal < rightVal-real ) 	return values[i-1];
		else								return values[i];
	}
}
